/***********************************************************************/
/*                                                                     */
/*                              Pio (C++)                              */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2005-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Estelle Marchand <Estelle.Marchand@inria.fr>                       */
/*                                                                     */
/***********************************************************************/

/* $Id: send.cpp,v 1.9 2014/01/10 19:22:09 fclement Exp $ */

#include "boolean.hpp"
#include "send.hpp"
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////


namespace io{
  //FOR DEFINITION OF send_lexem:
  
  // Lexem::                    (* those are roughly speaking C lexems *)
  //   | '<' Integer '>' '\n' String
  //   | Integer
  //   | Float
  //   | '(' Float ", " Float ", " Float ')'
  
  template<class dataType> void send_type(ostream& ob,const dataType & s){
    ob<<type(s,SEND)<<"\n";
  }
  //
  void send_value(ostream& ob,  const string& s){
    string s1(s);
    for(unsigned int i=s.size();i>0;i--){
      if (s1[i-1]<32 || s1[i-1]==34 || s1[i-1]==92 || s1[i-1]>126){
        short int a = s1[i-1];
        if(a<0) a+= 256;
        string escape_char = "\\";
        div_t temp = div(a,100);
        escape_char += ('0' + temp.quot );
        temp = div(temp.rem,10);
        escape_char += ('0' + temp.quot );
        escape_char += ('0' + temp.rem );
        s1.replace(i-1,1,escape_char);
      } 
    }
    ob<<'<'<<s.size()<<">\n"<<'\"'<<s1<<'\"';
  }
  //
  void send_value(ostream& ob,  const boolean& b){
    if(b.get_value()){
      ob<<"true";
    }else{
      ob<<"false";
    }
  }
  //
  void send_value(ostream& ob,  const int& i){
    ob<<i;
  }
  //
  void send_value(ostream& ob, const double& f){
    if(get_binary()){
      ob<<"&<8>";
      double* onedouble = new double(f);
      char* eightchar = 
        reinterpret_cast<char*> (onedouble);
      for(int i=0;i<8;i++)
        ob<<eightchar[i];
      delete(onedouble);
    }
    else{
      ob.width(get_width());
      ob.precision(get_prec());
      ob<<scientific<<right<<f;
    }
  }
  //
  template<class dataType,int n> void send_value(ostream& ob, const Tuple<dataType,n>& f){
    ob<<"(";
    for(int i=0;i<n-1;i++){
      send_value(ob,f[i]);
      ob<<", ";
    }
    send_value(ob,f[n-1]);
    ob<<")";
  }
  //
  template<class dataType1,class dataType2> void send_value(ostream& ob, const  pair<dataType1,dataType2>& f){
    ob<<"(";
    send_value(ob,f.first);
    ob<<", ";
    send_value(ob,f.second);
    ob<<")";
  }
  //
  template<class dataType> void send_lexem(ostream& ob, const  dataType& l){
    send_type(ob,l);
    send_value(ob,l);
    ob<<";\n";
  }
  //
  
  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////
  
  
  //FOR DEFINITION OF send_vector:

  template<class dataType> void send_type(ostream& ob,const vector<dataType>& v){
    dataType aal;
    string fmt=type(aal,SEND);
    ob<<"[1"<<fmt<<"1]\n";
  }
  
  template<class dataType> void send_items(ostream& ob,const vector<dataType>& v){
    unsigned int i;
    for(i=0;i<v.size();i++){
      send_value(ob,v[i]);
      ob<<";\n" ;
    }
  }
  
  template<class dataType> void send_vector(ostream& ob,const vector<dataType>& v){
    send_type(ob,v);
    ob<<"[1\n";
    ob<<"<"<<v.size()<<">\n";
    send_items(ob,v);
    ob<<"1];\n";
  }


  //function send_temp_items is used later: 

  template<class dataType> void send_temp_items(ostream& ob,int n,dataType*& d){
    for(int i=0;i<n;i++){
      send_value(ob,*d);
      ob<<";\n" ;
      d++;
    }
  }

  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////
  
  
  //SEND
  
  void w_begin(ostream& ob){
    ob<<"begin\n";
  }
  void w_end(ostream& ob){
    ob<<"end\n\n"<<flush;
  }
  
  template<class dataType> void send(ostream& ob,const dataType& t){
    w_begin(ob);
    send_lexem(ob,t);
    w_end(ob);
  }
  template<class dataType> void send(ostream& ob,const vector<dataType>& t){
    w_begin(ob);
    send_vector(ob,t);
    w_end(ob);
  }
  

  //functions declared in the header file:
  void send_string (ostream& ob,const string& t){send(ob,t);}
  void send_boolean (ostream& ob,const boolean& b){send(ob,b);}
  void send_boolean_couple (ostream& ob,const boolean_couple& c){send(ob,c);}
  void send_boolean_triple (ostream& ob,const boolean_triple& c){send(ob,c);}
  void send_int (ostream& ob,const int& i){send(ob,i);}
  void send_int_couple (ostream& ob,const int_couple& c){send(ob,c);}
  void send_int_triple (ostream& ob,const int_triple& c){send(ob,c);}
  void send_float (ostream& ob,const  double& x){send(ob,x);}
  void send_float_couple (ostream& ob,const float_couple& x){send(ob,x);}
  void send_float_triple (ostream& ob,const float_triple& x){send(ob,x);}
  void send_int_float_couple (ostream& ob,const int_float_couple& x){send(ob,x);}
  void send_float_int_couple(ostream& ob,const float_int_couple& x){send(ob,x);}
  void send_int_string_couple(ostream& ob,const int_string_couple& x){send(ob,x);}
  void send_string_int_couple(ostream& ob,const string_int_couple& x){send(ob,x);}
  void send_string_float_couple(ostream& ob,const string_float_couple& x){send(ob,x);}
  void send_float_string_couple(ostream& ob,const float_string_couple& x){send(ob,x);}
  void send_string_vector (ostream& ob,const vector<string>& v){send(ob,v);}
  void send_boolean_vector (ostream& ob,const vector<boolean>& v){send(ob,v);}
  void send_boolean_couple_vector (ostream& ob,const vector<boolean_couple>& v){send(ob,v);}
  void send_boolean_triple_vector (ostream& ob,const vector<boolean_triple>& v){send(ob,v);}
  void send_int_vector (ostream& ob,const vector<int>& v){send(ob,v);}
  void send_int_couple_vector (ostream& ob,const vector<int_couple>& v){send(ob,v);}
  void send_int_triple_vector (ostream& ob,const vector<int_triple>& v){send(ob,v);}
  void send_float_vector (ostream& ob,const vector<double>& v){send(ob,v);}
  void send_float_couple_vector (ostream& ob,const vector<float_couple>& v){send(ob,v);}
  void send_float_triple_vector (ostream& ob,const vector<float_triple>& v){send(ob,v);}
  
  
  template<class dataType> void send_type(ostream& ob,dataType* & M){
    w_begin(ob);
    dataType a;
    string fmt=type(a,SEND);
    ob<<"[2"<<fmt<<"2]\n";
  }

  template<class dataType> void send_type(ostream& ob,dataType** & M){
    w_begin(ob);
    dataType a;
    string fmt=type(a,SEND);
    ob<<"[2"<<fmt<<"2]\n";
  }



  void init_send_float_matrix(ostream& ob,const int& m,const int& n,string layout){
    ob<<"[2\n";
    ob<<'<'<<m<<", "<<n<<">\n"<<layout<<"\n";
  }
  
  void end_send_float_matrix(ostream& ob){
    ob<<"2];\n";
    w_end(ob);
  }
  
  void send_loop_layout_same_layout(ostream& ob,const int& m,const int& n,double* & M){
    double* MM=M;
    for(int mm=0;mm<m;mm++){
      ob<<"[|\n";
      send_temp_items(ob,n,MM);
      ob<<"|];\n";
    }
  }

  void send_loop_layout_other_layout(ostream& ob,const int& m,const int& n,double* & M){
    for(int nn=0;nn<n;nn++){
      ob<<"[|\n";
      for(int mm=0;mm<m;mm++){
        send_value(ob,M[n*mm+nn]);ob<<";\n";
      }
      ob<<"|];\n";
    }
  }
  
  void send_row_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"C");
    send_loop_layout_same_layout(ob,m,n,M);
    end_send_float_matrix(ob);
  }
  
  void send_row_float_matrix(ostream& ob,const int& m,const int& n,double** & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"C");
    double** MM=M;
    for(int mm=0;mm<m;mm++){
      double* K=*MM;
      ob<<"[|\n";
      send_temp_items(ob,n,K);
      ob<<"|];\n";
      MM++;
    }
    end_send_float_matrix(ob);
  }
  
  void send_column_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"F");
    send_loop_layout_other_layout(ob,m,n,M);
    end_send_float_matrix(ob);
  }
  
  void send_column_float_matrix(ostream& ob,const int& m,const int& n,double** & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"F");
    for(int nn=0;nn<n;nn++){
      ob<<"[|\n";
      for(int mm=0;mm<m;mm++){
        send_value(ob,M[mm][nn]);ob<<";\n";
      }
      ob<<"|];\n";
    }
    end_send_float_matrix(ob);
  }
 
  void send_row_to_row_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_row_float_matrix(ob,m,n,M);
  }
  void send_row_to_column_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_column_float_matrix(ob,m,n,M);
  }

  void send_column_to_row_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"C");
    send_loop_layout_other_layout(ob,n,m,M);
    end_send_float_matrix(ob);
  }
  
  void send_column_to_column_float_matrix(ostream& ob,const int& m,const int& n,double* & M){
    send_type(ob,M);
    init_send_float_matrix(ob,m,n,"F");
    send_loop_layout_same_layout(ob,n,m,M);
    end_send_float_matrix(ob);
  }
  
  
 void send_float_array(ostream& ob,const int& n,double* & M){
   int m=1;
   send_row_float_matrix(ob,m,n,M);
 }

}
