(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: write_communication.ml,v 1.11 2014/01/10 19:22:10 fclement Exp $ *)

open Communication;;

let debug fun_name msg =
  Options_pio.debug (Printf.sprintf "Write_communication.%s" fun_name) msg
;;

let w_values ob vals =
  assert (debug "w_values" "start writing");
  Array.iter (Write_typed_value.write_typed_value ob) vals;
  assert (debug "w_values" "finish writing")
;;

let d_com_msg = function
  | Phrase vals ->
    Printf.sprintf "Phrase of length %d" (Array.length vals)
  | Task (task_name, vals) ->
    Printf.sprintf "Task %S of length %d" task_name (Array.length vals)
  | Result vals ->
    Printf.sprintf "Result of length %d" (Array.length vals)
  | Error vals ->
    Printf.sprintf "Error of length %d" (Array.length vals)
  | Service (service_name, vals) ->
    Printf.sprintf "Service %S of length %d"
        (string_of_service_name service_name) (Array.length vals)
;;

let write_com ob c =
  assert
    (debug "write_com"
       (Printf.sprintf "start writing %s" (d_com_msg c)));
  begin match c with
  | Phrase vals ->
    Printf.bprintf ob "%%p <%d>\n%a" (Array.length vals) w_values vals
  | Task (task_name, vals) ->
    Printf.bprintf ob "%%t <%d> %S\n%a"
      (Array.length vals) task_name w_values vals
  | Result vals ->
    Printf.bprintf ob "%%r <%d>\n%a" (Array.length vals) w_values vals
  | Error vals ->
    Printf.bprintf ob "%%e <%d>\n%a" (Array.length vals) w_values vals
  | Service (service_name, vals) ->
    Printf.bprintf ob "%%s <%d> %S\n%a"
      (Array.length vals) (string_of_service_name service_name) w_values vals
  end;
  assert (debug "write_com" "finish writing")
;;

let write_communication ob c =
  Printf.bprintf ob "(\n%a)\n\n" write_com c
;;
