(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parser.mli,v 1.5 2014/01/21 09:21:18 fclement Exp $ *)

(* Monadic parsers *)
(* Copyright (c) Quentin Carbonneaux 2011 *)

type ('a, 'b) parser = { parse : 'a -> ('b * 'a) option };;

val parse : ('a, 'b) parser -> 'a -> 'b option;;
val full_parse : ('a, 'b) parser -> 'a -> ('b * 'a) option;;

val return : 'b -> ('a, 'b) parser;;
val zero : ('a, 'b) parser;;
val bind :
  ('a, 'b) parser -> ('b -> ('a, 'c) parser) -> ('a, 'c) parser;;
val plus : ('a, 'b) parser -> ('a, 'b) parser -> ('a, 'b) parser;;
val ( >>= ) :
  ('a, 'b) parser -> ('b -> ('a, 'c) parser) -> ('a, 'c) parser;;
val ( >> ) : ('a, 'b) parser -> ('a, 'c) parser -> ('a, 'c) parser;;
val ( +++ ) : ('a, 'b) parser -> ('a, 'b) parser -> ('a, 'b) parser;;

val liftM : ('a -> 'b) -> ('c, 'a) parser -> ('c, 'b) parser;;

val any : (('a, 'b) parser) list -> ('a, 'b) parser;;
val many : ('a, 'b) parser -> ('a, 'b list) parser;;
val many1 : ('a, 'b) parser -> ('a, 'b list) parser;;
val many_till :
  ('a, 'b) parser -> ('a, 'c) parser -> ('a, 'c list) parser;;
val times : int -> ('a, 'b) parser -> ('a, 'b list) parser;;
val sep_by1 :
  ('a, unit) parser -> ('a, 'b) parser -> ('a, 'b list) parser;;
val less_than : int -> ('a, 'b) parser -> ('a, 'b list) parser;;
val maybe : ('a, 'b) parser -> ('a, 'b option) parser;;
