(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2012 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: cbits.ml,v 1.12 2012/09/17 11:36:41 fclement Exp $ *)

(* IO with c programs prime_gen and prime_check *)

open Pool;;

let id x = x
let mkpair x y = (x, y)

let make_prime () =
  let local_primes =
    create
      (Options.get_pool_directory ())
      (Options.get_temporary_directory ())
      (Options.get_pool_size ())
  in
  let bin = ref None and cout = ref None in
  fun x ->
    try find local_primes x with Not_found ->
      let ib, oc = Uunix.spawn (Options.get_prime_gen_path ()) bin cout in
      Printf.fprintf oc "%d\n%!" (Options.get_factor_size ());
      try
        let prime, n_mods =
          Scanf.bscanf ib " prime = %s nmods = %d mods = " mkpair
        in
        let rec scan_mods i =
          if i = 0 then [] else
          Scanf.bscanf ib " %x;" id :: scan_mods (i - 1) in
        let mods = Array.of_list (scan_mods n_mods) in
        add local_primes x (prime, mods); (prime, mods)
      with
        Scanf.Scan_failure msg ->
          Printf.eprintf "! prime_gen is generating spurious data (%s).\n" msg;
          exit 2

(* let prime_check_intf local_primes = *)
(*   let bin = ref None and cout = ref None in *)
(*   fun factors cmd k -> *)
(*     if factors = [] then invalid_arg "empty factor list" else *)
(*     let ib, oc = Uunix.spawn (Options.get_prime_check_path ()) bin cout in *)
(*     let nfactors = List.length factors in *)
(*     let factors = List.map (fun x -> fst (find local_primes x)) factors in *)
(*     let req = List.fold_left *)
(*         (fun out x -> Printf.sprintf "%s\n%s" x out) "" factors in *)
(*     Printf.fprintf oc "nfactors = %d\n%s%s\n%!" nfactors req cmd; *)
(*     k ib *)

let prime_check_intf _ =
  let bin = ref None and cout = ref None in
  let prime_check_cmd =
    Printf.sprintf "%s %s %d"
      (Options.get_prime_check_path ())
      (Options.get_pool_directory ())
      (Options.get_pool_size ()) in
  fun factors cmd k ->
    if factors = [] then invalid_arg "empty factor list" else
    let ib, oc = Uunix.spawn prime_check_cmd bin cout in
    let nfactors = List.length factors in
    let req = List.fold_left
        (fun out x -> Printf.sprintf "%d; %s" x out) "" factors in
    Printf.fprintf oc "nfactors = %d\n%s\n%s\n%!" nfactors req cmd;
    k ib

let make_product local_pool =
  let intf = prime_check_intf local_pool in
  fun factors ->
    intf factors "m" (fun ib ->
      Scanf.bscanf ib " %s" id)

let check_product local_pool =
  let intf = prime_check_intf local_pool in
  fun factors ->
    intf factors "c" (fun ib ->
      Scanf.bscanf ib " %b" id)

(* keep this in sync with cbits/primes.h *)
let small_primes = [|
  0x0003; 0x0005; 0x0007; 0x000b; 0x000d;
  0x0011; 0x0013; 0x0017; 0x001d; 0x001f;
  0x0025; 0x0029; 0x002b; 0x002f; 0x0035;
  0x003b; 0x003d; 0x0043; 0x0047; 0x0049;
  0x004f; 0x0053; 0x0059; 0x0061; 0x0065;
  0x0067; 0x006b; 0x006d; 0x0071; 0x007f;
  0x0083; 0x0089; 0x008b; 0x0095; 0x0097;
  0x009d; 0x00a3; 0x00a7; 0x00ad; 0x00b3;
  0x00b5; 0x00bf; 0x00c1; 0x00c5; 0x00c7;
  0x00d3; 0x00df; 0x00e3; 0x00e5; 0x00e9;
  0x00ef; 0x00f1; 0x00fb; 0x0101; 0x0107;
  0x010d; 0x010f; 0x0115; 0x0119; 0x011b;
  0x0125; 0x0133; 0x0137; 0x0139; 0x013d;
  0x014b; 0x0151; 0x015b; 0x015d; 0x0161;
  0x0167; 0x016f; 0x0175; 0x017b; 0x017f;
  0x0185; 0x018d; 0x0191; 0x0199; 0x01a3;
  0x01a5; 0x01af; 0x01b1; 0x01b7; 0x01bb;
  0x01c1; 0x01c9; 0x01cd; 0x01cf; 0x01d3;
  0x01df; 0x01e7; 0x01eb; 0x01f3; 0x01f7;
  0x01fd; 0x0209; 0x020b; 0x021d; 0x0223;
  0x022d; 0x0233; 0x0239; 0x023b; 0x0241;
  0x024b; 0x0251; 0x0257; 0x0259; 0x025f;
  0x0265; 0x0269; 0x026b; 0x0277; 0x0281;
  0x0283; 0x0287; 0x028d; 0x0293; 0x0295;
  0x02a1; 0x02a5; 0x02ab; 0x02b3; 0x02bd;
  0x02c5; 0x02cf; 0x02d7; 0x02dd; 0x02e3;
  0x02e7; 0x02ef; 0x02f5; 0x02f9; 0x0301;
  0x0305; 0x0313; 0x031d; 0x0329; 0x032b;
  0x0335; 0x0337; 0x033b; 0x033d; 0x0347;
  0x0355; 0x0359; 0x035b; 0x035f; 0x036d;
  0x0371; 0x0373; 0x0377; 0x038b; 0x038f;
  0x0397; 0x03a1; 0x03a9; 0x03ad; 0x03b3;
  0x03b9; 0x03c7; 0x03cb; 0x03d1; 0x03d7;
  0x03df; 0x03e5; 0x03f1; 0x03f5; 0x03fb;
  0x03fd; 0x0407; 0x0409; 0x040f; 0x0419;
  0x041b; 0x0425; 0x0427; 0x042d; 0x043f;
  0x0443; 0x0445; 0x0449; 0x044f; 0x0455;
  0x045d; 0x0463; 0x0469; 0x047f; 0x0481;
  0x048b; 0x0493; 0x049d; 0x04a3; 0x04a9;
  0x04b1; 0x04bd; 0x04c1; 0x04c7; 0x04cd;
  0x04cf; 0x04d5; 0x04e1; 0x04eb; 0x04fd;
  0x04ff; 0x0503; 0x0509; 0x050b; 0x0511;
  0x0515; 0x0517; 0x051b; 0x0527; 0x0529;
  0x052f; 0x0551; 0x0557; 0x055d; 0x0565;
  0x0577; 0x0581; 0x058f; 0x0593; 0x0595;
  0x0599; 0x059f; 0x05a7; 0x05ab; 0x05ad;
  0x05b3; 0x05bf; 0x05c9; 0x05cb; 0x05cf;
  0x05d1; 0x05d5; 0x05db; 0x05e7; 0x05f3;
  0x05fb; 0x0607; 0x060d; 0x0611; 0x0617;
  0x061f; 0x0623; 0x062b; 0x062f; 0x063d;
  0x0641; 0x0647; 0x0649; 0x064d; 0x0653;
  0x0655; 0x065b; 0x0665; 0x0679; 0x067f;
  0x0683; 0x0685; 0x069d; 0x06a1; 0x06a3;
  0x06ad; 0x06b9; 0x06bb; 0x06c5; 0x06cd;
  0x06d3; 0x06d9; 0x06df; 0x06f1; 0x06f7;
  0x06fb; 0x06fd; 0x0709; 0x0713; 0x071f;
  0x0727; 0x0737; 0x0745; 0x074b; 0x074f;
  0x0751; 0x0755; 0x0757; 0x0761; 0x076d;
  0x0773; 0x0779; 0x078b; 0x078d; 0x079d;
  0x079f; 0x07b5; 0x07bb; 0x07c3; 0x07c9;
  0x07cd; 0x07cf; 0x07d3; 0x07db; 0x07e1;
  0x07eb; 0x07ed; 0x07f7; 0x0805; 0x080f;
  0x0815; 0x0821; 0x0823; 0x0827; 0x0829;
  0x0833; 0x083f; 0x0841; 0x0851; 0x0853;
  0x0859; 0x085d; 0x085f; 0x0869; 0x0871;
  0x0883; 0x089b; 0x089f; 0x08a5; 0x08ad;
  0x08bd; 0x08bf; 0x08c3; 0x08cb; 0x08db;
  0x08dd; 0x08e1; 0x08e9; 0x08ef; 0x08f5;
  0x08f9; 0x0905; 0x0907; 0x091d; 0x0923;
  0x0925; 0x092b; 0x092f; 0x0935; 0x0943;
|]
