/*
  Quentin Carbonneaux <qcarbonneaux@gmail.com>
  GPG Keyserver: pgp.mit.edu
  Key: C4D31021

  bignums, x86 optimized library dealing with
  arbitrary big positive numbers.

 */

#include "bignums.h"
#ifdef USE_ASM
/* This is a gcc extension */
/* #warning code is very specific to x86 platforms (USE_ASM enabled) */
#else
#include <string.h>
#endif

/* ************************************************ */

#define max(a,b) (((a)<(b))?(b):(a))
#define min(a,b) (((a)<(b))?(a):(b))

#define likely(x) __builtin_expect(x,1)
#define unlikely(x) __builtin_expect(x,0)

#define get_msb(msb,n) {                            \
    int32_t ___x; msb=0;                            \
    for(___x=(n).size/W_SIZE-1;___x>=0;___x--){     \
      if((n).buffer[___x]) {                        \
        msb=(___x+1)*W_SIZE-1;                      \
        while (!(((n).buffer[___x]>>(msb--))&0x1)); \
        msb++;                                      \
        break;                                      \
      }                                             \
    }                                               \
  }                                                 \

#define get_lsb(lsb,n) {                            \
    int32_t ___x; lsb=(n).size;                     \
    for(___x=0;___x<(n).size/W_SIZE;___x++){        \
      if((n).buffer[___x]) {                        \
        lsb=___x*W_SIZE;                            \
        while (!(((n).buffer[___x]>>(lsb++))&0x1)); \
        lsb--;                                      \
        break;                                      \
      }                                             \
    }                                               \
  }

/* ************************************************ */

UINT b_overflow=0;

/* ************************************************ */

UINT b_add_h(UINT *, UINT *, UINT *, uint32_t);
UINT b_multaddw_h(UINT *, UINT *, UINT, uint32_t);
void b_multw_h(UINT *, UINT *, UINT, uint32_t);
void b_shrw_h(UINT *, uint32_t);

/* ************************************************ */

struct b_int
b_alloc (uint32_t nbits) {
  struct b_int ret;

  ret.size = (nbits/W_SIZE)*W_SIZE;
  ret.buffer = xalloc(nbits/W_SIZE*sizeof(UINT));
  ret.neg = 0;

  return ret;
}

void
b_free (struct b_int n) {
  free (n.buffer);
}

void
b_zero (struct b_int n) {
#ifdef USE_ASM
  __asm__ volatile
    (
     "push %0 \n\t"
     "push %1 \n\t"
     "movl %%ecx, %%eax \n\t"
     "shr $1, %1 \n\t"
     "testl $1, %%eax \n\t"
     "jz 1f \n\t"
     "inc %1 \n\t"
#ifdef ARCH64
     "lea -8(%0), %0 \n\t"
     "jmp 2f \n\t"
     "1:"
     "movq $0, (%0) \n\t"
     "2:"
     "movq $0, 8(%0) \n\t"
     "lea 16(%0), %0 \n\t"
#else
     "lea -4(%0), %0 \n\t"
     "jmp 2f \n\t"
     "1:"
     "movl $0, (%0) \n\t"
     "2:"
     "movl $0, 4(%0) \n\t"
     "lea 8(%0), %0 \n\t"
#endif
     "dec %1 \n\t"
     "jnz 1b \n\t"
     "pop %1 \n\t"
     "pop %0 \n\t"
     : : "b"(n.buffer), "c"((SZ_CAST)(n.size/W_SIZE))
     : "%eax"
     );
#else  /* USE_ASM */
  memset(n.buffer, '\0', (n.size/W_SIZE)*(sizeof (UINT)));
#endif  /* not USE_ASM */
}

void
b_copy (struct b_int out, struct b_int in) {
#ifdef USE_ASM
  __asm__ volatile
    (
     "push %1 \n\t"
     "push %2 \n\t"
#ifdef ARCH64
     "mov %0, %%rcx \n\t"
     "sub $8, %1 \n\t"
     "sub $8, %2 \n\t"
     "1:"
     "mov (%1, %%rcx, 8), %%rax \n\t"
     "mov %%rax, (%2, %%rcx, 8) \n\t"
     "dec %%rcx \n\t"
#else
     "mov %0, %%ecx \n\t"
     "sub $4, %1 \n\t"
     "sub $4, %2 \n\t"
     "1:"
     "movl (%1, %%ecx, 4), %%eax \n\t"
     "movl %%eax, (%2, %%ecx, 4) \n\t"
     "dec %%ecx \n\t"
#endif
     "jnz 1b \n\t"
     "pop %2 \n\t"
     "pop %1 \n\t"
     : : "r"((SZ_CAST)min(out.size/W_SIZE, in.size/W_SIZE)),
       "r"(in.buffer), "r"(out.buffer)
#ifdef ARCH64
     : "rax", "%rcx"
#else
     : "eax", "%ecx"
#endif
     );
#else  /* USE_ASM */
  memcpy(out.buffer, in.buffer,
         (min(out.size/W_SIZE, in.size/W_SIZE))*(sizeof (UINT)));
#endif  /* not USE_ASM */
}

void
b_zcopy (struct b_int out, struct b_int in) {
#ifdef USE_ASM
  __asm__ volatile
    (
     "push %0 \n\t"
     "push %2 \n\t"
     "push %3 \n\t"
#ifdef ARCH64
     "sub $8, %2 \n\t"
     "sub $8, %3 \n\t"
     "cmp %0, %1 \n\t"
     "jge 1f \n\t"
     "2:"
     "movq $0, (%2, %0, 8) \n\t"
#else
     "sub $4, %2 \n\t"
     "sub $4, %3 \n\t"
     "cmp %0, %1 \n\t"
     "jge 1f \n\t"
     "2:"
     "movl $0, (%2, %0, 4) \n\t"
#endif
     "dec %0 \n\t"
     "cmp %0, %1 \n\t"
     "jne 2b \n\t"
     "1:"
#ifdef ARCH64
     "mov (%3, %0, 8), %%rax \n\t"
     "mov %%rax, (%2, %0, 8) \n\t"
#else
     "movl (%3, %0, 4), %%eax \n\t"
     "movl %%eax, (%2, %0, 4) \n\t"
#endif
     "dec %0 \n\t"
     "jnz 1b \n\t"
     "pop %3 \n\t"
     "pop %2 \n\t"
     "pop %0 \n\t"
     : : "b"((SZ_CAST)out.size/W_SIZE), "c"((SZ_CAST)in.size/W_SIZE),
       "r"(out.buffer), "r"(in.buffer)
#ifdef ARCH64
     : "%rax"
#else
     : "%eax"
#endif
     );
#elif 0  /* USE_ASM */
  register int32_t i;

  if (out.size > in.size) {
    for (i=out.size/W_SIZE-1; i > in.size/W_SIZE-1; i--)
      out.buffer[i] = 0x0;
    out.size = in.size;
  }
  for (i=0; i<out.size/W_SIZE; i++) {
    out.buffer[i] = in.buffer[i];
  }
#else
  b_zero(out);
  b_copy(out, in);
#endif  /* not USE_ASM */
}

#if 0
void
b_swap (struct b_int a, struct b_int b) {
  int32_t i;

  for (i=0; i<min(a.size/W_SIZE, b.size/W_SIZE); i++) {
    __asm__(
            "movl (%0), %%eax \n\t"
            "xchgl %%eax, (%1) \n\t"
            "movl %%eax, (%0) \n\t"
            : : "r"(a.buffer+i), "r"(b.buffer+i)
            : "%eax"
            );
  }
}
#endif

void
b_add (struct b_int out, struct b_int a, struct b_int b) {
#ifdef USE_ASM
  assert(a.size == b.size && a.size/W_SIZE > 0);

  b_overflow = b_add_h(out.buffer, a.buffer, b.buffer, a.size/W_SIZE);
#else  /* USE_ASM */
  register int32_t i;
  unsigned char cf=0;

  assert(a.size == b.size);

  for (i=0; i<out.size/W_SIZE; i++) {
    UINT l = a.buffer [i];
    unsigned char cf0 = 0;

    if (unlikely(cf && (l = l + cf) == 0))
      cf0 = 1;
    cf = 0;
    l += b.buffer[i];
    if (l < b.buffer[i])
      cf = 1;
    cf |= cf0;
    (out.buffer)[i] = l;
  }

  b_overflow = cf;
#endif  /* not USE_ASM */
}

void
b_mult (struct b_int out, struct b_int a, struct b_int b) {
#ifdef USE_ASM
  struct b_int tmp;
  register int32_t i;

  assert(a.size + b.size <= out.size);

  tmp = b_alloc(out.size);
  b_zero(out);

  for (i=0; i<b.size/W_SIZE; i++) {
    if (b.buffer[i] > 1) {
      b_multw(tmp, b.buffer[i], a);
      b_shl(tmp, tmp, i*W_SIZE);
      b_add(out, out, tmp);
    } else if (b.buffer[i] == 1) {
      b_shl(tmp, a, i*W_SIZE);
      b_add(out, out, tmp);
    }
  }
  b_free(tmp);
#else  /* USE_ASM */
  struct b_int tmp;
  register int32_t bit;

  assert(a.size + b.size <= out.size);

  tmp = b_alloc(out.size);
  b_zero(out);

  for (bit=0; bit<b.size; bit++){
    if ((b.buffer[bit/W_SIZE]>>(bit%W_SIZE))&0x1) {
      /* Shift 'a' to 'tmp' and add it to the result */
      b_shl (tmp, a, bit);
      b_add(out, out, tmp);
    }
  }
  b_free(tmp);
#endif  /* not USE_ASM */
}

void
b_multw (struct b_int out, UINT a, struct b_int b) {
#ifdef USE_ASM
  assert(out.size >= (b.size + W_SIZE) && b.size > 1);

  b_zero(out);
  b_multw_h(out.buffer, b.buffer, a, b.size/W_SIZE);
#else  /* USE_ASM */
  struct b_int tmp;

  tmp.size = W_SIZE;
  tmp.buffer = &a;
  tmp.neg = 0;
  b_mult(out, b, tmp);
#endif  /* not USE_ASM */
}

/* out = a * b + c */
/* optimized for out == c (in the sense of pointers) */
void
b_multaddw (struct b_int out, UINT a, struct b_int b,
            struct b_int c) {
#ifdef USE_ASM
  assert(out.size >= (b.size + W_SIZE) && b.size > 1 && c.size > b.size);

  if (likely(out.buffer == c.buffer)) {
    b_overflow = b_multaddw_h(out.buffer, b.buffer, a, b.size/W_SIZE);
  }
  else {
    struct b_int tmp = b_alloc(c.size);

    b_copy(tmp, c);
    b_overflow = b_multaddw_h(tmp.buffer, b.buffer, a, b.size/W_SIZE);
    b_zcopy(out, tmp);
    b_free(tmp);
  }
  if (b_overflow && b.size + W_SIZE < out.size) { /* optimize here */
    struct b_int carry = b_alloc(out.size);

    b_zero(carry);
    carry.buffer[b.size/W_SIZE + 1] = 0x1;
    b_add(out, out, carry);
    b_free(carry);
  }
#else  /* USE_ASM */
  struct b_int tmp = b_alloc(c.size);

  assert(out.size > b.size && b.size > 1 && c.size > b.size);

  b_multw(tmp, a, b);
  b_add(out, tmp, c);
  b_free(tmp);
#endif /* not USE_ASM */
}

int
b_is_zero (struct b_int n) {
#ifdef USE_ASM
  int out;

  __asm__ volatile
    (
     "movl $1, %0 \n\t"
     "push %2 \n\t"
     "1:"
     "cmp $0, (%2) \n\t"
     "jne 2f \n\t"
#ifdef ARCH64
     "add $8, %2 \n\t"
#else
     "add $4, %2 \n\t"
#endif
     "decl %1 \n\t"
     "jnz 1b \n\t"
     "jmp 3f \n\t"
     "2:"
     "movl $0, %0 \n\t"
     "3:"
     "pop %2 \n\t"
     : "=m"(out) : "c"(n.size/W_SIZE), "r"(n.buffer)
     );
  return out;
#else  /* USE_ASM */
  register int32_t i;

  for (i=0; i<n.size/W_SIZE; i++)
    if (n.buffer[i] != 0)
      return 0;

  return 1;
#endif  /* not USE_ASM */
}

int
b_is_one (struct b_int n) {
#ifdef USE_ASM
  int out;

  __asm__ volatile
    (
     "movl $1, %0 \n\t"
     "push %2 \n\t"
     "cmp $1, (%2) \n\t"
     "jne 2f \n\t"
#ifdef ARCH64
     "add $8, %2 \n\t"
#else
     "add $4, %2 \n\t"
#endif
     "decl %1 \n\t"
     "jz 3f \n\t"
     "1:"
     "cmp $0, (%2) \n\t"
     "jne 2f \n\t"
#ifdef ARCH64
     "add $8, %2 \n\t"
#else
     "add $4, %2 \n\t"
#endif
     "decl %1 \n\t"
     "jnz 1b \n\t"
     "jmp 3f \n\t"
     "2:"
     "movl $0, %0 \n\t"
     "3:"
     "pop %2 \n\t"
     : "=m"(out) : "c"(n.size/W_SIZE), "r"(n.buffer)
     );
  return out;
#else  /* USE_ASM */
  register int32_t i;

  if (n.buffer[0] != 1)
    return 0;
  for (i=1; i<n.size/W_SIZE; i++)
    if (n.buffer[i] != 0)
      return 0;
  return 1;
#endif  /* not USE_ASM */
}

int
b_gt (struct b_int a, struct b_int b) {
  register int32_t i;

  if (a.size > b.size) {
    while (a.size != b.size) {
      if (a.buffer[a.size/W_SIZE-1])
        return 1;
      a.size -= W_SIZE;
    }
  }
  if (a.size < b.size) {
    while (a.size != b.size) {
      if (b.buffer[b.size/W_SIZE-1])
        return 0;
      b.size -= W_SIZE;
    }
  }

  for (i=a.size/W_SIZE-1; i>=0; i--) {
    if ((a.buffer[i] > b.buffer[i])) {
      return 1;
    } else {
      if (a.buffer[i] != b.buffer[i])
        return 0;
    }
  }
  return 0;
}

int
b_geqt (struct b_int a, struct b_int b) {
  register int32_t i;

  if (a.size > b.size) {
    while (a.size != b.size) {
      if (a.buffer[a.size/W_SIZE-1])
        return 1;
      a.size -= W_SIZE;
    }
  }
  if (a.size < b.size) {
    while (a.size != b.size) {
      if (b.buffer[b.size/W_SIZE-1])
        return 0;
      b.size -= W_SIZE;
    }
  }

  for (i=a.size/W_SIZE-1; i>=0; i--) {
    if ((a.buffer[i] > b.buffer[i])) {
      return 1;
    } else {
      if (a.buffer[i] != b.buffer[i])
        return 0;
    }
  }
  return 1;
}

void
b_sub (struct b_int out, struct b_int a, struct b_int b) {
  int32_t i;
  unsigned char cf=0;

  assert(a.size == b.size);
  assert(b_geqt(a,b));

  for (i=0; i<out.size/W_SIZE; i++) {
    UINT l = a.buffer [i];
#ifdef USE_ASM
    __asm__ volatile
      (
#ifdef ARCH64
       "movzb %2, %%rax \n\t"
       "sub %%rax, %1 \n\t"
#else
       "movzbl %2, %%eax \n\t"
       "subl %%eax, %1 \n\t"
#endif
       "setc %0 \n\t"
       "sub %4, %1 \n\t"
       "setc %%al \n\t"
       "orb %%al, %0 \n\t"
       : "=m"(cf), "=m"(l)
       : "m"(cf), "m"(l), "r"(b.buffer[i])
       : "%eax"
       );
#else  /* USE_ASM */
    unsigned char cf0 = 0;

    if (unlikely(cf && (l = l - cf) == ~((UINT)0x0)))
      cf0 = 1;
    cf = 0;
    if (l - b.buffer[i] > l)
      cf = 1;
    l -= b.buffer[i];
    cf |= cf0;
#endif  /* not USE_ASM */
    (out.buffer)[i] = l;
  }

  b_overflow = cf;              /* Must always be 0 */
}

void
b_adds (struct b_int *out, struct b_int a, struct b_int b) {
  struct b_int tmp = b_alloc(max(a.size, b.size));

  if (a.neg ^ b.neg) {
    if (b_gt(a,b)) {
      b_sub(tmp, a, b);
      tmp.neg = a.neg;
    } else {
      b_sub(tmp, b, a);
      tmp.neg = b.neg;
    }
  } else {
    b_add(tmp, a, b);
    tmp.neg = a.neg;
  }
  b_zcopy(*out, tmp);
  out->neg = tmp.neg;
  b_free(tmp);
}

void
b_subs (struct b_int *out, struct b_int a, struct b_int b) {
  struct b_int tmp = b_alloc(b.size);

  b_copy(tmp, b);
  tmp.neg = !b.neg;

  b_adds(out, a, tmp);

  b_free(tmp);
}

void
b_shl (struct b_int out, struct b_int n, int32_t shift) {
  struct b_int tmp;
  const int32_t sh=shift%W_SIZE;
#ifdef USE_ASM
  unsigned char write_over;
#endif  /* USE_ASM */
  int32_t i;

  tmp = b_alloc(out.size);
#ifdef USE_ASM

  b_zero(tmp);
  i = min(n.size/W_SIZE, tmp.size/W_SIZE - shift/W_SIZE);
  write_over = (n.size<(tmp.size-shift))?'\1':'\0';

  __asm__ volatile
    (
     "push %[tmp] \n\t"
     "push %[n] \n\t"

     "test %[count], %[count] \n\t"
     "jz 4f \n\t"
#ifdef ARCH64
     "sub $8, %[n] \n\t"
     "mov (%[n], %[count], 8), %%rax \n\t"
     "xor %%rbx, %%rbx \n\t"
     "shld %%cl, %%rax, %%rbx \n\t"
     "push %%rax \n\t"
     "movb %[wover], %%al \n\t"
     "testb %%al, %%al \n\t"
     "jz 5f \n\t"
     "mov %%rbx, (%[tmp], %[count], 8) \n\t"
     "5:"
     "pop %%rax \n\t"
     "dec %[count] \n\t"
     "jz 3f \n\t"
     "test $1, %[count] \n\t"
     "jnz 2f \n\t"
     "mov %%rax, %%rbx \n\t"

     "1:"
     "mov (%[n], %[count], 8), %%rax \n\t"
     "shld %%cl, %%rax, %%rbx \n\t"
     "mov %%rbx, (%[tmp], %[count], 8) \n\t"
     "dec %[count] \n\t"
     "2:"
     "mov (%[n], %[count], 8), %%rbx \n\t"
     "shld %%cl, %%rbx, %%rax \n\t"
     "mov %%rax, (%[tmp], %[count], 8) \n\t"
     "dec %[count] \n\t"
     "jnz 1b \n\t"

     "shl %%cl, %%rbx \n\t"
     "mov %%rbx, (%[tmp]) \n\t"
     "jmp 4f \n\t"

     "3:"
     "shl %%cl, %%rax \n\t"
     "mov %%rax, (%[tmp]) \n\t"
#else
     "sub $4, %[n] \n\t"
     "movl (%[n], %[count], 4), %%eax \n\t"
     "xorl %%ebx, %%ebx \n\t"
     "shldl %%cl, %%eax, %%ebx \n\t"
     "push %%eax \n\t"
     "movb %[wover], %%al \n\t"
     "testb %%al, %%al \n\t"
     "jz 5f \n\t"
     "movl %%ebx, (%[tmp], %[count], 4) \n\t"
     "5:"
     "pop %%eax \n\t"
     "dec %[count] \n\t"
     "jz 3f \n\t"
     "test $1, %[count] \n\t"
     "jnz 2f \n\t"
     "movl %%eax, %%ebx \n\t"

     "1:"
     "movl (%[n], %[count], 4), %%eax \n\t"
     "shldl %%cl, %%eax, %%ebx \n\t"
     "movl %%ebx, (%[tmp], %[count], 4) \n\t"
     "dec %[count] \n\t"
     "2:"
     "movl (%[n], %[count], 4), %%ebx \n\t"
     "shldl %%cl, %%ebx, %%eax \n\t"
     "movl %%eax, (%[tmp], %[count], 4) \n\t"
     "dec %[count] \n\t"
     "jnz 1b \n\t"

     "shll %%cl, %%ebx \n\t"
     "movl %%ebx, (%[tmp]) \n\t"
     "jmp 4f \n\t"

     "3:"
     "shll %%cl, %%eax \n\t"
     "movl %%eax, (%[tmp]) \n\t"
#endif
     "4:"
     "pop %[n] \n\t"
     "pop %[tmp] \n\t"
     : : [wover]"m"(write_over),
       [tmp]"r"(tmp.buffer+(shift/W_SIZE)),
       [n]"r"(n.buffer),
       [sh]"c"((uint32_t)sh%W_SIZE),
       [count]"r"((SZ_CAST)i)
#ifdef ARCH64
     : "%rax", "%rbx"
#else
     : "%eax", "%ebx"
#endif
     );
#else  /* USE_ASM */
  b_zero(tmp);

  if (sh) {
    for (i=0; i<min(n.size/W_SIZE,out.size/W_SIZE-shift/W_SIZE-1); i++) {
      tmp.buffer[i+shift/W_SIZE] |= (n.buffer[i]<<sh);
      tmp.buffer[i+1+shift/W_SIZE] |= (n.buffer[i]>>(W_SIZE-sh));
    }
    if (unlikely((i==out.size/W_SIZE-shift/W_SIZE-1) && (i+1<n.size/W_SIZE))) {
      tmp.buffer[i+shift/W_SIZE] |= (n.buffer[i]<<sh);
    }
  } else {
    for (i=0; i<min(n.size/W_SIZE,out.size/W_SIZE-shift/W_SIZE); i++) {
      tmp.buffer[i+shift/W_SIZE] = (n.buffer[i]);
    }
  }
#endif  /* not USE_ASM */
  b_zcopy(out, tmp);
  b_free(tmp);
}

void
b_shr (struct b_int out, struct b_int n, int32_t shift) {
  struct b_int tmp;
  const int32_t sh=shift%W_SIZE;
  int32_t i;

  assert(shift<=n.size);

  tmp = b_alloc(out.size); b_zero(tmp);

  tmp.buffer[0] = (n.buffer[shift/W_SIZE]>>sh);

  if (sh) {
    for (i=shift/W_SIZE+1; i<n.size/W_SIZE; i++) {
      tmp.buffer[i-shift/W_SIZE] |= (n.buffer[i]>>sh);
      tmp.buffer[i-1-shift/W_SIZE] |= (n.buffer[i]<<(W_SIZE-sh));
    }
  } else {
    for (i=shift/W_SIZE+1; i<n.size/W_SIZE;i++) {
      tmp.buffer[i-shift/W_SIZE] = (n.buffer[i]);
    }
  }
  b_zcopy(out, tmp);
  b_free(tmp);
}

void
b_shrw (struct b_int out, struct b_int n) {
  if (unlikely(out.buffer != n.buffer))
    b_zcopy(out, n);
  b_shrw_h(out.buffer, out.size/W_SIZE);
}

/* Quite faster than the first version ...*/
void
b_mod (struct b_int out, struct b_int n, struct b_int mod) {
  int32_t msb_n;
  int32_t msb_mod;
  struct b_int tmp = b_alloc (n.size);
  struct b_int tmp_out = b_alloc (n.size);

  b_zero(tmp_out); b_copy(tmp_out, n);
  get_msb(msb_mod, mod);

  while (b_geqt(tmp_out, mod)) {
    get_msb(msb_n, tmp_out);
    b_shl(tmp,mod,max(msb_n-msb_mod-1,0));
    b_sub(tmp_out, tmp_out, tmp);
  }
  b_zcopy(out, tmp_out);
  assert(b_gt(mod, out));
  b_free(tmp);
  b_free(tmp_out);
}

#ifdef B_EUCLIDE_GCD
void
b_gcd (struct b_int out, struct b_int a, struct b_int b) {
  struct b_int tmp_a = b_alloc (a.size);
  struct b_int tmp_b = b_alloc (b.size);
  struct b_int *p_a, *p_b, *p_0;

  p_a = &tmp_a;
  p_b = &tmp_b;

  if (likely(b_geqt(b,a))) {
    b_copy(tmp_a, a);
    b_copy(tmp_b, b);
  } else {
    b_copy(tmp_a, b);
    b_copy(tmp_b, a);
  }

  while (!b_is_zero(*p_a)) {
    b_mod(*p_b, *p_b, *p_a);
    p_0 = p_a;
    p_a = p_b;
    p_b = p_0;
  }
  b_zcopy(out, *p_b);
  b_free(tmp_a);
  b_free(tmp_b);
}
#else  /* B_EUCLIDE_GCD */
void
b_gcd (struct b_int out, struct b_int a, struct b_int b) {
  struct b_int tmp_a = b_alloc(a.size);
  struct b_int tmp_b = b_alloc(b.size);
  struct b_int *p_a, *p_b;
  uint32_t g, lsb_a, lsb_b;

  get_lsb(lsb_a, a);
  get_lsb(lsb_b, b);

  if (unlikely(lsb_a == a.size)) {
    b_zero(out);
    b_copy(out, b);
    goto SAFE_QUIT;
  }
  if (unlikely(lsb_b == b.size)) {
    b_zero(out);
    b_copy(out, a);
    goto SAFE_QUIT;
  }

  b_copy(tmp_a, a);
  b_copy(tmp_b, b);
  g = min(lsb_a, lsb_b);

  if (b_geqt(tmp_a, tmp_b)) {
    p_a = &tmp_b;
    p_b = &tmp_a;
  } else {
    p_a = &tmp_a;
    p_b = &tmp_b;
  }

  while (!b_is_zero(*p_a)) {
    get_lsb(lsb_a, *p_a);
    get_lsb(lsb_b, *p_b);

    if (lsb_a)
      b_shr(*p_a, *p_a, lsb_a);
    if (lsb_b)
      b_shr(*p_b, *p_b, lsb_b);

    if (b_geqt(*p_a, *p_b)) {
      b_sub(*p_a, *p_a, *p_b);
      b_shr(*p_a, *p_a, 1);
    } else {
      b_sub(*p_b, *p_b, *p_a);
      b_shr(*p_b, *p_b, 1);
    }
  }

  b_shl(out, *p_b, g);

 SAFE_QUIT:
  b_free(tmp_a);
  b_free(tmp_b);
}

void
b_egcd (struct b_int out, struct b_int *a, struct b_int *b,
        struct b_int x, struct b_int y) {
  uint32_t g, lsb_x, lsb_y;
  struct b_int A = b_alloc(max(x.size, y.size));
  struct b_int B = b_alloc(max(x.size, y.size));
  struct b_int C = b_alloc(max(x.size, y.size));
  struct b_int D = b_alloc(max(x.size, y.size));
  struct b_int u = b_alloc(x.size), v = b_alloc(y.size);
  struct b_int old_x = b_alloc(x.size), old_y = b_alloc(y.size);

  get_lsb(lsb_x, x);
  get_lsb(lsb_y, y);
  g = min(lsb_x, lsb_y);

  b_copy(u, x); b_copy(v, y);
  b_shr(u, u, g);
  b_shr(v, v, g);

  b_copy(old_x, u);
  b_copy(old_y, v);
  b_zero(A); b_zero(B);
  b_zero(C); b_zero(D);
  A.buffer[0] = 0x1;
  D.buffer[0] = 0x1;

  while (!b_is_zero(u)) {
    while (!(u.buffer[0] & 0x1)) {
      b_shr(u, u, 1);
      if ((A.buffer[0] | B.buffer[0]) & 0x1) {
        b_adds(&A, A, old_y);
        b_subs(&B, B, old_x);
      }
      b_shr(A, A, 1);
      b_shr(B, B, 1);
    }
    while (!(v.buffer[0] & 0x1)) {
      b_shr(v, v, 1);
      if ((C.buffer[0] | D.buffer[0]) & 0x1) {
        b_adds(&C, C, old_y);
        b_subs(&D, D, old_x);
      }
      b_shr(C, C, 1);
      b_shr(D, D, 1);
    }
    if (b_geqt(u, v)) {
      b_sub(u, u, v);
      b_subs(&A, A, C);
      b_subs(&B, B, D);
    } else {
      b_sub(v, v, u);
      b_subs(&C, C, A);
      b_subs(&D, D, B);
    }
  }

  if (a) {
    b_zero(*a);
    b_copy(*a, C); a->neg = C.neg;
  }
  if (b) {
    b_zero(*b);
    b_copy(*b, D); b->neg = D.neg;
  }
  b_shl(v, v, g);
  b_zcopy(out, v);

 /* SAFE_QUIT: */
  b_free(old_y); b_free(old_x);
  b_free(v); b_free(u);
  b_free(D); b_free(C);
  b_free(B); b_free(A);
}
#endif  /* not B_EUCLIDE_GCD */

void
b_invmod (struct b_int out, struct b_int a, struct b_int m) {
  struct b_int tmp_out = b_alloc(max(a.size, m.size));
  struct b_int gcd = b_alloc(max(a.size, m.size));

  b_egcd(gcd, &tmp_out, NULL, a, m);
  assert(gcd.buffer[0] == 0x1);
  while (tmp_out.neg)
    b_adds(&tmp_out, tmp_out, m);
  b_mod(tmp_out, tmp_out, m);
  b_zcopy(out, tmp_out);
  b_free(tmp_out);
  b_free(gcd);
}

void
b_mgm_free (struct b_mgm_ctx ctx) {
  b_free(ctx.r2);
  b_free(ctx.r);
}

struct b_mgm_ctx
b_mgm_init (struct b_int m) {
  const uint32_t p = m.size + W_SIZE;
  struct b_int R = b_alloc(p + W_SIZE);
  struct b_int R2 = b_alloc(2*p + W_SIZE);
  struct b_int tmp1 = b_alloc(m.size);
  struct b_int tmp2 = b_alloc(2*p + W_SIZE);
  struct b_int b = b_alloc(m.size);
  struct b_mgm_ctx out;

  out.p = p;

  assert(m.buffer[0] & 0x1);
  b_zero(b); b.buffer[1] = 0x1;
  b_invmod(tmp1, m, b);
  b_sub(tmp1, b, tmp1);
  out.n = tmp1.buffer[0];

  /* b_zero(tmp2); */
  b_zcopy(tmp2, m);

  tmp2.size = p + W_SIZE;
  R.buffer[0] = 0x1;
  b_shl(R, R, p);
  b_mod(R, R, tmp2);
  R.size = p;           /* shrink it */

  tmp2.size = 2*p + W_SIZE;
  R2.buffer[0] = 0x1;
  b_shl(R2, R2, 2*p);
  b_mod(R2, R2, tmp2);
  R2.size = p;          /* shrink it */

  out.r = R;
  out.r2 = R2;

  b_free(tmp1);
  b_free(tmp2);
  b_free(b);

  return out;
}

void
b_mgm_mult (struct b_int out, struct b_int a, struct b_int b,
            struct b_int m, struct b_mgm_ctx context) {
  struct b_int A = b_alloc(context.p + W_SIZE);
  struct b_int tmp = b_alloc(context.p + W_SIZE);
  UINT q;
  uint32_t i;

  assert(a.size == context.p);
  assert(b.size == context.p && b.buffer[b.size/W_SIZE-1] == 0x0);
  assert(out.size == context.p);

  b_zero(A);

  for (i=0; i<context.p/W_SIZE; i++) {
    q = (A.buffer[0] + a.buffer[i]*b.buffer[0]) * context.n;
    /* b_multw(tmp, a.buffer[i], b); */
    /* b_add(A, A, tmp); */
    /* b_multw(tmp, q, m); */
    /* b_add(A, A, tmp); */
    b_multaddw(A, a.buffer[i], b, A);
    b_multaddw(A, q, m, A);
    b_shrw(A, A);
  }

  b_zcopy(out, A);
  b_free(A);
  b_free(tmp);
}

void
b_mgm_powm (struct b_int out, struct b_int g, struct b_int e,
           struct b_int m, struct b_mgm_ctx context) {
  register int32_t bit;
  struct b_int A = b_alloc(context.p);
  struct b_int x = b_alloc(context.p);
  struct b_int one = b_alloc(context.p);

  b_zero(one);
  one.buffer[0] = 0x1;

  b_mgm_mult(x, g, context.r2, m, context);
  b_copy(A, context.r);

  get_msb(bit, e);
  for (; bit>=0; bit--) {
    b_mgm_mult(A, A, A, m, context);
    if (e.buffer[bit/W_SIZE]>>(bit%W_SIZE)&0x1)
      b_mgm_mult(A, x, A, m, context);
  }
  b_mgm_mult(A, A, one, m, context);

  b_zcopy(out, A);
  A.size -= W_SIZE; assert(b_gt(m, A));
  b_free(one);
  b_free(x);
  b_free(A);
}

int
b_miller_rabin (struct b_int n, uint32_t t,
                void (*rand_gen) (UINT *, uint32_t, int),
                struct b_mgm_ctx *ctxop) {
  uint32_t lsb, msb;
  struct b_int n1 = b_alloc(n.size);
  struct b_int y = b_alloc(n.size + W_SIZE);
  struct b_int y2 = b_alloc(2*(n.size + W_SIZE));
  struct b_int one = b_alloc(n.size + W_SIZE);
  struct b_int r = b_alloc(n.size);
  struct b_int a = b_alloc(n.size + W_SIZE);
  struct b_int pow = b_alloc(n.size + W_SIZE);
  struct b_mgm_ctx ctx = ctxop ? *ctxop : b_mgm_init(n);
  uint32_t i, j;
  int ret = -1;

  t = max(t, 5);                /* make sure t is big enough */
  if (!(n.buffer[0]&0x1))
    goto SAFE_QUIT;
  b_zero(one); one.buffer[0] = 0x1;
  one.size = n.size;
  b_copy(r, n);
  b_sub(r, r, one);
  get_lsb(lsb, r);
  b_copy(n1, r);
  b_shr(r, r, lsb);
  get_msb(msb, r);
  one.size = n.size + W_SIZE;

  for (i=0; i<t; i++) {
    do {
      rand_gen(a.buffer, msb-1, 0);
    } while (b_is_zero(a));
    b_mgm_powm(y, a, r, n, ctx);
    if ((!b_is_one(y)) && b_gt(n1, y)) {
      for (j=1; j<lsb && b_gt(n1, y); j++) {
#if 0
        b_mgm_mult(y, y, ctx.r2, n, ctx);
        b_mgm_mult(y, y, y, n, ctx);
        b_mgm_mult(y, y, one, n, ctx);
#else
        b_mult(y2, y, y);
        b_mod(y, y2, n);
#endif
        assert(b_gt(n, y));
        if (b_is_one(y))
          goto SAFE_QUIT;
      }
      if (b_gt(n1, y))
        goto SAFE_QUIT;
    }
  }

  ret = 0;
 SAFE_QUIT:
  b_free(pow);
  b_free(a);
  b_free(one);
  b_free(r);
  b_free(y2);
  b_free(y);
  b_free(n1);
  if (!ctxop) b_mgm_free(ctx);
  return ret;
}

/* ************************************************ */

struct b_int
gen_prime (int32_t nbits, void (*rand) (UINT *, uint32_t, int),
              int rand_lvl) {
  int32_t psize = nbits;
  struct b_int p;
  struct b_int tmp;
  struct b_int mod;
  struct b_int one;
  struct b_int two;
  int32_t i, j;
  int isprime = 0;
  uint32_t *mods;
  int32_t msb;

  for (; psize%W_SIZE; psize++);
  p = b_alloc(psize);
  b_zero(p);
  tmp = b_alloc(psize);
  b_zero(tmp);
  mod = b_alloc(psize);
  two = b_alloc(psize);
  b_zero(two); two.buffer[0] = 0x2;
  one = b_alloc(psize);
  b_zero(one); one.buffer[0] = 0x1;
  mods = xalloc((sizeof b_primes/sizeof b_primes[0])*sizeof(uint32_t));

  while (!isprime) {
    rand(p.buffer, nbits, rand_lvl);
    p.buffer[(nbits-1)/W_SIZE] |= ((UINT)0x1)<<((nbits-1)%W_SIZE);
    p.buffer[0] |= 0x1;

    for (i=0; i<(sizeof b_primes / sizeof b_primes[0]); i++) {
      tmp.buffer[0] = b_primes[i];
      b_mod(mod, p, tmp);
      mods[i] = mod.buffer[0];
    }
    for /*(i=0; i<100000; i++)*/ (;;) {
      isprime = 1;
      for (j=0; j<(sizeof b_primes / sizeof b_primes[0]); j++) {
        if (mods[j] == 0x0) {
          isprime = 0;
          break;
        }
      }
      if (isprime) {
        struct b_mgm_ctx ctx = b_mgm_init(p);
        { /* simple fermat test */
          struct b_int p1 = b_alloc(p.size);
          struct b_int a = b_alloc(p.size+W_SIZE);
          struct b_int pow = b_alloc(p.size+W_SIZE);

          b_sub(p1, p, one);
          b_zcopy(a, two);
          b_mgm_powm(pow, a, p1, p, ctx);
          if (!b_is_one(pow))
            isprime=0;
          b_free(pow);
          b_free(a);
          b_free(p1);
        }
        if (isprime)
          if (b_miller_rabin(p, 10, rand, &ctx) == 0)
            break;
        b_mgm_free(ctx);
      }

      b_add(p, p, two);
      get_msb(msb, p);
      if (msb > nbits || b_overflow) { /* overflow */
        isprime = 0;
        break;
      }

      for (j=0; j<(sizeof b_primes / sizeof b_primes[0]); j++) {
        mods[j] += 2;
        while (mods[j] >= b_primes[j])
          mods[j] -= b_primes[j];
      }
    }
  }
  free(mods);
  b_free(two);
  b_free(one);
  b_free(mod);
  b_free(tmp);
  return p;
}

static int
next_comb (char combs[], int n) {
  int i = 0, n1 = 0;

  while (!((n1 += (int)combs[i]) && !(combs[i]))) {
    if (++i == n)
      return 0;
  }
  n1--;
  combs[i--] = '\1';

  while (i >= 0) {
    if (i >= n1)
      combs[i--] = '\0';
    else
      combs[i--] = '\1';
  }
  return 1;
}

struct b_int
b_gen_prime (struct b_int **fact, int *nfact, int32_t nbits, int32_t qbits,
             void (*rand) (UINT *, uint32_t, int),
             int rand_lvl) {
  int n = 1;
  int32_t i, j;
  int res, isprime;
  struct b_int *factors;
  struct b_int p;
  struct b_int tmp;
  struct b_int one;
  struct b_int two;
  struct b_int modprime;
  struct b_int mod;
  int nfactors;
  char *perms;
  uint32_t **mods;
  uint32_t *index;

  assert((nbits > 50) && (qbits < nbits));

  if (!qbits)
    qbits = nbits/3;
  while (nbits/n >= qbits)
    n++;
  n--;
  /* all factors, except 2 have at least qbits bits */

  nfactors = 4*n;               /* experimental, for now */
  nfactors = max(nfactors,30);
  factors = xalloc(nfactors*(sizeof(struct b_int)));
  perms = xalloc(nfactors*(sizeof(char)));
  mods = xalloc(nfactors*(sizeof(uint32_t *)));
  index = xalloc(n*sizeof(uint32_t));
  for (i=0; i<nfactors; i++) {
    factors[i].buffer = NULL;
    perms[i] = ((i<n)?'\1':'\0');
    mods[i] = NULL;
  }

  p = b_alloc(nbits + nfactors + W_SIZE + 1);
  tmp = b_alloc(p.size + qbits + W_SIZE);
  one = b_alloc(p.size);
  two = b_alloc(p.size + W_SIZE);
  mod = b_alloc(p.size);
  modprime = b_alloc(p.size);
  b_zero(one); b_zero(modprime); b_zero(two);
  one.buffer[0] = 0x1;
  two.buffer[0] = 0x2;

 next_try:
  res = 1;
  isprime = 0;
  for (;;) {
    int k = 0;
    b_zero(p);
    p.buffer[0] = 0x2;
    for (i=0; i<nfactors; i++) {
      if (perms[i]) {
        if (!factors[i].buffer) {
          factors[i] = gen_prime(qbits, rand, rand_lvl);
          mods[i] = xalloc((sizeof b_primes / sizeof b_primes[0])*
                           sizeof(uint32_t));
          for (j=0; j<(sizeof b_primes / sizeof b_primes[0]); j++) {
            modprime.buffer[0] = b_primes[j];
            b_mod(mod, factors[i], modprime);
            mods[i][j] = mod.buffer[0];
          }
        }
        index[k++] = i;
      }
    }
    /* check if moduli are ok */
    /* XXX heavily rely on the fact that small primes are smaller than
       half of a word */
    isprime = 1;
    for (j=0; j<(sizeof b_primes / sizeof b_primes[0]); j++) {
      uint32_t mod = 2;
      for (i=0; i<n; i++) {
        mod *= mods[index[i]][j];
        mod %= b_primes[j];
      }
      mod |= 1;
      mod -= b_primes[j];
      if (!mod) {
        isprime = 0;
        break;
      }
    }

    if (isprime) {              /* generate it */
      for (i=0; i<n; i++) {
        b_mult(tmp, p, factors[index[i]]);
#ifndef NDEBUG
        {
          int32_t msb;
          get_msb(msb, tmp);
          assert(msb < p.size);
        }
#endif
        b_copy(p, tmp);
      }
      /* p = 2*q1*q2*...*qn + 1 */
      p.buffer[0] |= 0x1;
    }
    if (isprime) {
      /* simple fermat test */
      struct b_int pow = b_alloc(p.size+W_SIZE);
      struct b_int p1 = b_alloc(p.size);
      struct b_mgm_ctx ctx;
      /* int32_t msbp; */

      /* get_msb(msbp, p); */
      b_sub(p1, p, one);
      ctx = b_mgm_init(p);
      b_mgm_powm(pow, two, p1, p, ctx);
      b_free(p1);
      if (b_is_one(pow)) {      /* run big tests */
        if (b_miller_rabin(p, 3, rand, &ctx) == 0) {
          isprime = 1;
        }
        else
          isprime = 0;
      } else
        isprime = 0;
      b_mgm_free(ctx);
      b_free(pow);
    }

    if (isprime)
      break;

    if (res == 0) {
      for (i=0; i<nfactors; i++) {
        b_free(factors[i]);
        factors[i].buffer = NULL;
        perms[i] = ((i<n)?'\1':'\0');
        free(mods[i]);
      }
      goto next_try;
    }
    /* generate the next combination of primes */
    res = next_comb(perms, nfactors);
  }
  /* got it */
  *nfact = n;
  *fact = xalloc(n*(sizeof(struct b_int)));
  for (i=0; i<n; i++) {
    (*fact)[i] = b_alloc(qbits+W_SIZE);
    b_zero((*fact)[i]);
  }
  for (i=0, j=0; i<nfactors; i++) {
    if (perms[i]) {
      b_copy((*fact)[j++], factors[i]);
    }
    if (factors[i].buffer) {
      b_free(factors[i]);
      free(mods[i]);
    }
  }
  free(index);
  free(mods);
  free(factors);
  free(perms);
  b_free(mod);
  b_free(modprime);
  b_free(two);
  b_free(one);
  b_free(tmp);
  return p;
}

void
b_sprint (char *str, struct b_int n) {
  int32_t i;

  for (i=(n.size/W_SIZE)-1; i>=0; i--) {
#ifdef ARCH64
    sprintf (str+(n.size/W_SIZE-1-i)*16, "%016" PRIx64, n.buffer[i]);
#else
    sprintf (str+(n.size/W_SIZE-1-i)*8, "%08" PRIx32, n.buffer[i]);
#endif
  }
}

/* ************************************************ */

#ifdef TEST
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define INITRANDOM fdrandom = open("/dev/urandom", O_RDONLY);
static int fdrandom;

static void
test_rand (UINT *buffer, uint32_t nbits, int sec) {
  int32_t nwords;

  if (nbits%W_SIZE)
    nwords = (nbits+W_SIZE)/W_SIZE;
  else
    nwords = nbits/W_SIZE;

  (void)read(fdrandom, buffer, nwords*sizeof(UINT));
  buffer[nwords-1] &= (~(UINT)0)>>(W_SIZE - (nbits%W_SIZE));
}

int
main ()
{
#if T==8
  struct b_int p;
  char s[10024];
  struct b_int *facts;
  int nfacts;

  INITRANDOM;

  p = b_gen_prime(&facts, &nfacts, 4096, 1024, test_rand, 0);
  b_sprint(s, p);

  /* Test 8 : generate a 2048 bits strong prime number */
  printf("generated prime : %s\n", s);

#endif

#if T==7
  struct b_int p;
  char s[1024];

  INITRANDOM;

  p = gen_prime(1024, test_rand, 0);
  b_sprint(s, p);

  /* Test 7 : generate a 1024 bits prime number */
  printf("generated prime : %s\n", s);
  /* The result must be a fat prime number, check it with */
  /* the GMP or openssl or libgcrypt or what you want. */

  /* The security of this generator is really _poor_ since */
  /* it is based on the random function. */
#endif

#if T==6
  char s1[1024], s2[1024];
#ifdef ARCH64
  struct b_int n1 = b_alloc(2*W_SIZE);
  struct b_int n2 = b_alloc(2*W_SIZE);

  INITRANDOM;

  n1.buffer[0] = UINT64_C(0x9bb3920ef5e958b9);
  n1.buffer[1] = UINT64_C(0x2);
  n2.buffer[1] = UINT64_C(0x87b57be17);
  n2.buffer[0] = UINT64_C(0xf0ecdbf18A227bd9);

#else
  struct b_int n1 = b_alloc(3*W_SIZE);
  struct b_int n2 = b_alloc(4*W_SIZE);

  INITRANDOM;

  n1.buffer[0] = UINT32_C(0xf5e958b9);
  n1.buffer[1] = UINT32_C(0x9bb3920e);
  n1.buffer[2] = UINT32_C(0x2);
  n2.buffer[3] = UINT32_C(0x8);
  n2.buffer[2] = UINT32_C(0x7b57be17);
  n2.buffer[1] = UINT32_C(0xf0ecdbf1);
  n2.buffer[0] = UINT32_C(0x8A227bd9);
#endif
  b_sprint(s1, n1); b_sprint(s2, n2);

  /* Test 6 : check if primes are primes */
  printf("MillerRabin(%s, %d) = %d\n",
         s1, 5,
         b_miller_rabin(n1, 5, test_rand, 0));

  printf("MillerRabin(%s, %d) = %d\n",
         s2, 50,
         b_miller_rabin(n2, 50, test_rand, 0));

  /* The result of this test must show that the two */
  /* numbers are primes (ret=0) : */
  /* MillerRabin( 000000029bb3920ef5e958b9 , 5) = 0 */
  /* MillerRabin( 000000087b57be17f0ecdbf18a227bd9 , 50) = 0 */
#endif

#if T==5
  struct b_mgm_ctx mgm;
#ifdef ARCH64
  struct b_int m = b_alloc(2*W_SIZE);
  struct b_int g = b_alloc(3*W_SIZE);
  struct b_int e = b_alloc(2*W_SIZE);
  struct b_int out = b_alloc(3*W_SIZE);

  b_zero(m);
  b_zero(g); b_zero(e);

  m.buffer[0] = UINT64_C(199); m.buffer[1] = UINT64_C(0x8000000000000000);
  g.buffer[0] = UINT64_C(67); g.buffer[1] = UINT64_C(0x8000000000000000);
  e.buffer[0] = UINT64_C(354);

  mgm = b_mgm_init(m);
  b_mgm_powm(out, g, e, m, mgm);

  /* Test 5 : montgomery exponentiation */
  printf("montgomery exponentiation g^e is %" PRIu64 " %" PRIu64 "\n",
         out.buffer[1], out.buffer[0]);
  /* The result of this test must be */
  /* montgomery exponentiation g^e is */
  /* 8395749123063678083 3186573207560592007 */
#else
  struct b_int m = b_alloc(2*W_SIZE);
  struct b_int g = b_alloc(3*W_SIZE);
  struct b_int e = b_alloc(2*W_SIZE);
  struct b_int out = b_alloc(3*W_SIZE);

  b_zero(m);
  b_zero(g); b_zero(e);

  m.buffer[0] = UINT32_C(199); m.buffer[1] = UINT32_C(0x80000000);
  g.buffer[0] = UINT32_C(67); g.buffer[1] = UINT32_C(0x80000000);
  e.buffer[0] = UINT32_C(354);

  mgm = b_mgm_init(m);
  b_mgm_powm(out, g, e, m, mgm);

  /* Test 5 : montgomery exponentiation */
  printf("montgomery exponentiation g^e is %" PRIu32 " %" PRIu32 "\n",
         out.buffer[1], out.buffer[0]);
  /* The result of this test must be */
  /* montgomery exponentiation g^e is 564205810 4147828058 */
#endif
#endif

#if T==4
  struct b_mgm_ctx mgm;
  struct b_int m = b_alloc(2*W_SIZE);
  struct b_int a = b_alloc(3*W_SIZE);
  struct b_int b = b_alloc(3*W_SIZE);
  struct b_int c = b_alloc(3*W_SIZE);

  b_zero(a); b_zero(b); b_zero(c);
  b_zero(m);

#ifdef ARCH64
  m.buffer[0] = UINT64_C(0x00000001); m.buffer[1] = UINT64_C(1);
  a.buffer[0] = UINT64_C(0x100000017);
  b.buffer[0] = UINT64_C(0x2000000b4);
#else
  m.buffer[1] = UINT32_C(1); m.buffer[0] = UINT32_C(1);
  a.buffer[0] = UINT32_C(23); a.buffer[1] = UINT32_C(1);
  b.buffer[0] = UINT32_C(180); b.buffer[1] = UINT32_C(2);
#endif

  mgm = b_mgm_init(m);
  b_mgm_mult(c, a, b, m, mgm);

  /* Test 4 : montgomery multiplication */
#ifdef ARCH64
  printf("montgomery product axb is %" PRIu64 "\n", c.buffer[0]);
#else
  printf("montgomery product axb is %" PRIu32 "\n", c.buffer[0]);
#endif
  /* The result of this test must be */
  /* 32bits: montgomery product axb is 4294963381 */
  /* 64bits: montgomery product axb is 18446743103046938583 */
#endif

#if T==3
  struct b_mgm_ctx mgm;
  struct b_int modulus = b_alloc(2*W_SIZE);

  modulus.buffer[0] = 8867;
  mgm = b_mgm_init(modulus);

  /* Test 3 : test b_egcd again and b_mgm_init (one part only) */
  printf("b_mgm_init: n = %u, n' = %u\n",
         (unsigned)modulus.buffer[0], mgm.n);

  /* The result of the test 3 must be */
  /* b_mgm_init: n = 8867, n' = 4188888821 */

  /* Or, if you set modulus.buffer[0] = 973 : */
  /* b_mgm_init: n = 973, n' = 4025704187 */
#endif

#if T==2
  struct b_int a,b,gcd,x,y;
  char *sa = malloc(1024);
  char *sb = malloc(1024);
  char *sgcd = malloc(1024);
  char *sx = malloc(1024);
  char *sy = malloc(1024);

#ifdef ARCH64
  a = b_alloc (W_SIZE*2);
  b = b_alloc (W_SIZE*2);
  gcd = b_alloc (W_SIZE*2);
  x = b_alloc (W_SIZE*2);
  y = b_alloc (W_SIZE*2);
#else
  a = b_alloc (W_SIZE*3);
  b = b_alloc (W_SIZE*3);
  gcd = b_alloc (W_SIZE*3);
  x = b_alloc (W_SIZE*3);
  y = b_alloc (W_SIZE*3);
#endif

  b_zero(a);
  b_zero(b);

  /* Test 2: Extended gcd of two numbers */
#ifdef ARCH64
  a.buffer[0]=UINT64_C(0x40bc08d3a3);
  b.buffer[0]=UINT64_C(0xdf425e2ae77051c3);
  b.buffer[1]=UINT64_C(0xa4);
#else
  a.buffer[0]=UINT32_C(0xbc08d3a3);
  a.buffer[1]=UINT32_C(0x40);
  b.buffer[0]=UINT32_C(0xe77051c3);
  b.buffer[1]=UINT32_C(0xdf425e2a);
  b.buffer[2]=UINT32_C(0xa4);
#endif
  b_egcd(gcd, &x, &y, a, b);

  b_sprint(sa, a); b_sprint(sb, b);
  b_sprint(sx, x); b_sprint(sy, y);
  b_sprint(sgcd, gcd);

  printf(" %s * %s\n+-\n %s * %s\n=\n  %s nx=%d ny=%d\n",
         sa, sx, sb, sy, sgcd, x.neg, y.neg);

  b_zero(a);
  b_zero(b);

  /* This is the correct output for test 2 */
  /*   00000000  00000040  bc08d3a3  *  00000017  5f0b4e7b  03670c45  */
  /* +- */
  /*   000000a4  df425e2a  e77051c3  *  00000000  00000009  2d2696c4  */
  /* = */
  /*    00000000  00000040  bc08d3a3  nx=0 ny=1 */
#endif

#if T==1
  struct b_int a,b,c;
  char *sa = malloc(1024);
  char *sb = malloc(1024);
  char *sc = malloc(1024);

  a = b_alloc (W_SIZE*3);
  b = b_alloc (W_SIZE*3);
  c = b_alloc (W_SIZE*3);

  b_zero(a);
  b_zero(b);

  /* Test 1 : gcd of these two numbers must be a */
#ifdef ARCH64
  a.buffer[0]=UINT64_C(0x40bc08d3a3);
  b.buffer[0]=UINT64_C(0xdf425e2ae77051c3);
  b.buffer[1]=UINT64_C(0xa4);
#else
  a.buffer[0]=UINT32_C(0xbc08d3a3);
  a.buffer[1]=UINT32_C(0x40);
  b.buffer[0]=UINT32_C(0xe77051c3);
  b.buffer[1]=UINT32_C(0xdf425e2a);
  b.buffer[2]=UINT32_C(0xa4);
#endif

  b_mod(c, b, a);

  b_sprint(sa, a); b_sprint(sb, b); b_sprint(sc, c);

  printf("%s\nrel\n%s\n=\n%s\n",sa,sb,sc);
  /*fprintf(stderr, "%s %llx * %s + pq", sb, 0xf00040050f000400, sa);*/
#endif

#if T==0
  struct b_int a, b, c;
  char sa[1024], sb[1024], sc[1024];

  a = b_alloc(W_SIZE*3);
  b = b_alloc(W_SIZE*5);
  c = b_alloc(W_SIZE*4);

  b_zero(a); b_zero(b); b_zero(c);

#ifdef ARCH64
  a.buffer[0]=UINT64_C(0xffffffffffffff56);
  a.buffer[2]=a.buffer[1]=UINT64_C(0xffffffffffffffff);
  c.buffer[0]=UINT64_C(0xffffffffffffffff);
  c.buffer[3]=UINT64_C(0xcf000000000000000);

  b_multaddw(b, UINT64_C(0xffffffffffffffff), a, c);
  b_sprint(sa, a); b_sprint(sb, b); b_sprint(sc, c);
  printf("%s\n%s\n%s\n",sa,sb,sc);

  /* The output must be : */
  /* ffffffffffffffffffffffffffffffffffffffffffffff56 */
  /* 0000000000000001effffffffffffffeffffffffffffffffffffffffffffff5700000000000000a9 */
  /* f00000000000000000000000000000000000000000000000ffffffffffffffff */
#else
  a.buffer[0]=UINT32_C(0xffffff56);
  a.buffer[2]=a.buffer[1]=UINT32_C(0xffffffff);
  c.buffer[0]=UINT32_C(0xffffffff);
  c.buffer[3]=UINT32_C(0xf0000000);

  b_multaddw(b, UINT32_C(0xffffffff), a, c);
  b_sprint(sa, a); b_sprint(sb, b); b_sprint(sc, c);
  printf("%s\n%s\n%s\n",sa,sb,sc);

  /* The ouput must be : */
  /* ffffffffffffffffffffff56 */
  /* 00000001effffffeffffffffffffff57000000a9 */
  /* f00000000000000000000000ffffffff */
#endif

#endif

  return 0;
}
#endif
