#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "bignums.h"

#define USE_INTERNAL_CACHE /* define this to use a cache (modify caml
                              accordingly) */

#define DEFAULT_BUF_SZ 4096

#define get_msb(msb,n) {                                        \
    int32_t ___x; msb=0;                                        \
    for(___x=(n).size/W_SIZE-1;___x>=0;___x--){                 \
      if((n).buffer[___x]) {                                    \
        msb=(___x+1)*W_SIZE-1;                                  \
        while (!(((n).buffer[___x]>>(msb--))&0x1));             \
        msb++;                                                  \
        break;                                                  \
      }                                                         \
    }                                                           \
  }                                                             \

int b_miller_rabin (struct b_int, uint32_t, void (*) (UINT *, uint32_t, int),
                    struct b_mgm_ctx *);
void b_mgm_powm (struct b_int, struct b_int, struct b_int, struct b_int,
                 struct b_mgm_ctx);
void b_mgm_mult (struct b_int, struct b_int, struct b_int, struct b_int,
                 struct b_mgm_ctx);
struct b_mgm_ctx b_mgm_init (struct b_int);
void b_mgm_free (struct b_mgm_ctx);
int b_is_one(struct b_int);

#ifdef USE_INTERNAL_CACHE
struct b_int **pool;
#endif

int fdrand;
char *pname;
char *tmpdir;
static unsigned ascii_to_hex[] = {
  0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
  0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
};

static char *
scan_word (UINT *n, char *s)
{
  int nsteps = W_SIZE / 4;
  char c;

  *n = 0;
  for (; (c = *s++) && nsteps--;) {
    *n <<= 4; *n |= (UINT)ascii_to_hex[(int)c];
  }
  return s - 1;
}

static struct b_int
scan_b_int (char *s, int sz)
{
  struct b_int ret;
  int nwords;

  for (; sz % W_SIZE; sz++);
  nwords = sz / W_SIZE;
  ret = b_alloc(sz);
  while (nwords--) {
    s = scan_word(ret.buffer + nwords, s);
  }
  return ret;
}

static void
die (char *s)
{
  fprintf(stderr, "%s dying -- %s\n", pname, s);
  exit(2);
}

#ifndef USE_INTERNAL_CACHE

static struct b_int
get_b_int (FILE *f)
{
  char *buf, *p;
  int bufsz = DEFAULT_BUF_SZ;
  int i = 0;
  struct b_int ret;
  if (!(buf = malloc(DEFAULT_BUF_SZ))) die("out of memory");
  do {
    if (i >= bufsz) {
      bufsz = 2 * bufsz; buf = realloc(buf, bufsz);
      if (!buf) die("out of memory");
    }
    if (0 == fread(buf + i, 1, 1, f)) die("premature end of input");
  } while (buf[i++] != '\n');
  i--;
  p = buf;
  for (; *buf == '0'; buf++, i--);
  if (i * 4 < W_SIZE) die("input factor too small");
  if (i % (W_SIZE/4)) {
    ret = scan_b_int(buf + (i % (W_SIZE/4)), 4 * i - W_SIZE);
    ret.buffer = realloc(ret.buffer, (1 + ret.size/W_SIZE) * sizeof(UINT));
    if (!ret.buffer)
      die("out of memory");
    ret.size += W_SIZE;
    buf[i % (W_SIZE / 4)] = '\0';
    scan_word(&ret.buffer[ret.size/W_SIZE - 1], buf);
  } else {
    ret = scan_b_int(buf, 4 * i);
  }
  free(p);
  return ret;
}

#else

static struct b_int
unpack_prime (unsigned int n)
{
  struct b_int ret;
  char *p;
  char fname[1024];
  FILE *caml_data;
  unsigned char block_tag, string_tag = '\0';
  unsigned long string_len;

  if (snprintf(fname, 1024, "%s/%d", tmpdir, n) >= 1024)
    die("aren't you kidding with your paths ?...");
  if (!(caml_data = fopen(fname, "r")))
    die("unable to open a prime file (sync error ?)");
  fseek(caml_data, 20, SEEK_SET); /* skip headers */
  (void)fread(&block_tag, 1, 1, caml_data);
  (void)fread(&string_tag, 1, 1, caml_data);
  if (string_tag != 0x09 && string_tag != 0x0a)
    die("invalid string tag");
  if (string_tag == 0x09) { /* STRING8  */
    unsigned char c;
    (void)fread(&c, 1, 1, caml_data);
    string_len = c;
  } else {                  /* STRING32 */
    unsigned char word32[4] = "    ";
    (void)fread(word32, 1, 4, caml_data);
    string_len =
      (word32[0] << 24) +
      (word32[1] << 16) +
      (word32[2] << 8)  +
      (word32[3]);
  }
  if (!(p = malloc(string_len+1)))
    die("out of memory");
  fread(p, 1, string_len, caml_data);
  p[string_len] = '\0';
  fclose(caml_data);
  if (string_len % (W_SIZE/4)) {
    ret = scan_b_int(p + (string_len % (W_SIZE/4)), 4 * string_len - W_SIZE);
    ret.buffer = realloc(ret.buffer, (1 + ret.size/W_SIZE) * sizeof(UINT));
    if (!ret.buffer)
      die("out of memory");
    ret.size += W_SIZE;
    p[string_len % (W_SIZE/4)] = '\0';
    scan_word(&ret.buffer[ret.size/W_SIZE - 1], p);
  } else {
    ret = scan_b_int(p, 4 * string_len);
  }
  free(p);
  return ret;
}

#endif

static void
xrand (UINT *buffer, uint32_t nbits, int lvl)
{
  int32_t nwords;

  if (nbits%W_SIZE)
    nwords = (nbits+W_SIZE)/W_SIZE;
  else
    nwords = nbits/W_SIZE;

  (void)read(fdrand, buffer, nwords*sizeof(UINT));
  buffer[nwords-1] &= (~(UINT)0)>>(W_SIZE - (nbits%W_SIZE));
}

int
main (int argc, char **argv)
{
  unsigned nfactors, psize, max_size = 0;
  unsigned i;
  char c;
  struct b_int *factors, p, tmp;

#ifdef USE_INTERNAL_CACHE
  unsigned int pool_size;
  unsigned int word;
#endif

  pname = argv[0];

#ifdef USE_INTERNAL_CACHE
  if (argc<3)
    die("usage : prime_check DIR POOL_SIZE");
  tmpdir = argv[1];
  pool_size = atoi(argv[2]);
  if (!pool_size)
    die("usage : prime_check DIR POOL_SIZE");
  if (!(pool = malloc(pool_size * sizeof(struct b_int *))))
    die("out of memory");
  for (i = 0; i < pool_size; i++)
    pool[i] = NULL;
#endif

  fdrand = open("/dev/urandom", O_RDONLY);
  while (1) {
    psize = 0;
#ifndef USE_INTERNAL_CACHE
    max_size = 0;
#endif
    if (scanf(" nfactors = %u ", &nfactors) < 1) break;
    if (nfactors == 0) break;

    factors = malloc(nfactors * sizeof(struct b_int));
    for (i = 0; i < nfactors; i++) {
#ifdef USE_INTERNAL_CACHE
      if (scanf(" %u; ", &word) < 1)
        die("not enough factors given");
      if (pool[word] == NULL) {
        struct b_int *prime = malloc(sizeof(struct b_int));
        if (!prime) die("out of memory");
        *prime = unpack_prime(word);
        pool[word] = prime;
        if (prime->size > max_size)
          max_size = prime->size;
      }
      factors[i] = *(pool[word]);
#else
      //char t[1000];
      factors[i] = get_b_int(stdin);
      if (factors[i].size > max_size) max_size = factors[i].size;
      //b_sprint(t, factors[i]);
      //printf("%s\n", t);
#endif
      psize += factors[i].size;
    }
    /* let's compute */
    p = b_alloc(psize + W_SIZE);
    b_zero(p); p.buffer[0] = 0x2;
    tmp = b_alloc(p.size + max_size);
    for (i = 0; i < nfactors; i++) {
      b_mult(tmp, p, factors[i]);
#ifndef NDEBUG
      {
        int32_t msb;
        get_msb(msb, tmp);
        assert(msb < p.size);
      }
#endif
      b_copy(p, tmp);
    }
    p.buffer[0] |= 0x1;
    b_free(tmp);
    if (0 == fread(&c, 1, 1, stdin)) die("premature end of input");
    if (c == 'm') {
      char out[p.size/4 + 1];
      b_sprint(out, p);
      printf("%s\n", out);
    }
    else { /* check primality */
      /* simple fermat test */
      int isprime;
      struct b_int pow = b_alloc(p.size + W_SIZE);
      struct b_int p1 = b_alloc(p.size);
      struct b_int two = b_alloc(p.size + W_SIZE);
      struct b_int one = b_alloc(p.size);
      struct b_mgm_ctx ctx;
      ctx = b_mgm_init(p);
      b_zero(two); b_zero(one); two.buffer[0] = 0x2; one.buffer[0] = 0x1;
      b_sub(p1, p, one);
      b_mgm_powm(pow, two, p1, p, ctx);
      b_free(p1);
      if (b_is_one(pow)) {      /* run big tests */
        if (b_miller_rabin(p, 3, xrand, &ctx) == 0) isprime = 1;
        else isprime = 0;
      } else isprime = 0;
      b_mgm_free(ctx);
      b_free(pow);
      b_free(one);
      b_free(two);
      if (isprime) { printf("true\n"); } else { printf("false\n"); }
    }
    fflush(stdout);
    b_free(p);
#ifndef USE_INTERNAL_CACHE
    for (i = 0; i < nfactors; i++)
      b_free(factors[i]);
#endif
    free(factors);
  }
#ifdef USE_INTERNAL_CACHE
  for (i = 0; i < pool_size; i++) {
    if (pool[i]) free(pool[i]);
  }
  free(pool);
#endif
  close(fdrand);
  return 0;
}
