(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: make_domain.ml,v 1.10 2014/01/21 09:21:16 fclement Exp $ *)

open Skl;;

type 'a borders = ('a border) list
and 'a border = (int * 'a);;

type ('a, 'b) worker_spec = ('a borders, 'a * 'b) Skl.t * int list;;

let from_Some o = match o with Some x -> x | None -> invalid_arg "from_Some";;

let sanity_check speca =
  let n = Array.length speca in
  let check_conn = List.for_all ((>) n) in
  let anding f = fun x y -> x && f y in
  Array.fold_left (anding (fun (_, l) -> check_conn l)) true speca
;;

let mux : (int list) array -> (('a * 'b) array, ('a borders) array) Skl.t =
  let mux_pure = fun conn_spec -> Sklstream.map (
    fun a -> Array.map (List.map (fun i -> (i, fst a.(i)))) conn_spec
  ) in
  ignore (Expor.register mux_pure);
  fun conn_spec ->
    atom (conn_spec, from_Some (Expor.is_registered mux_pure))
;;

let dmux : (int list) array -> (('a * 'b) array, ('a * 'b option) array) Skl.t =
  let dmux_pure = fun _ -> Sklstream.map (
    Array.map (fun (x, y) -> (x, Some y))
  ) in
  ignore (Expor.register dmux_pure);
  fun conn_spec ->
    atom (conn_spec, from_Some (Expor.is_registered dmux_pure))
;;

let ppcond : unit -> (('a * 'b option) array, 'b array) Skl.t =
  let ppcond_pure = fun () -> Sklstream.map (
    Array.map (fun (_, x) -> from_Some x)
  ) in
  ignore (Expor.register ppcond_pure);
  fun () -> atom ((), from_Some (Expor.is_registered ppcond_pure))
;;

let pre : unit -> ('a array, ('a * 'b option) array) Skl.t =
  let pre_pure = fun () -> Sklstream.map (
    Array.map (fun x -> (x, None))
  ) in
  ignore (Expor.register pre_pure);
  fun () -> atom ((), from_Some (Expor.is_registered pre_pure))
;;

let post : unit -> (('a * 'b option) array, ('a * 'b) array) Skl.t =
  let post_pure = fun () -> Sklstream.map (
    Array.map (fun (x, y) -> (x, from_Some y))
  ) in
  ignore (Expor.register post_pure);
  fun () -> atom ((), from_Some (Expor.is_registered post_pure))
;;

let make_make_domain speca pskl =
  if not (sanity_check speca)
  then invalid_arg "make_domain: wrong input specification" else
  let conn_spec = Array.map snd speca in
  pre () |||
  loop ( pskl
       , mux conn_spec ||| rails (Array.map fst speca) ||| dmux conn_spec) |||
  post ()
;;

let make_domain speca pskl = make_make_domain speca (ppcond () ||| pskl);;

let make_f_domain speca p =
  let pskl_reg =
    Expor.register (
      fun () -> Sklstream.map (
        fun a -> p (Array.map (fun (_, x) -> from_Some x) a)
      )
    ) in
  let pskl = atom ((), pskl_reg) in
  make_make_domain speca pskl
;;

let pre_opt : unit -> ('a array, ('a * 'b) array) Skl.t =
  let pre_opt_pure = fun () -> Sklstream.map (
    let magic_default = Obj.magic 0 in
    Array.map (fun x -> (x, magic_default))
  ) in
  ignore (Expor.register pre_opt_pure);
  fun () -> atom ((), from_Some (Expor.is_registered pre_opt_pure))
;;

let ppcond_opt : unit -> (('a * 'b) array, 'b array) Skl.t =
  let ppcond_opt_pure = fun () -> Sklstream.map (
    Array.map snd
  ) in
  ignore (Expor.register ppcond_opt_pure);
  fun () -> atom ((), from_Some (Expor.is_registered ppcond_opt_pure))
;;

let make_make_domain_opt speca pskl =
  if not (sanity_check speca)
  then invalid_arg "make_domain: wrong input specification" else
  let conn_spec = Array.map snd speca in
  pre_opt () |||
  loop ( pskl
       , mux conn_spec ||| rails (Array.map fst speca))
;;

let make_domain_opt speca pskl =
  make_make_domain_opt speca (ppcond_opt () ||| pskl)
;;

let make_f_domain_opt speca p =
  let pskl_reg =
    Expor.register (
      fun () -> Sklstream.map (
        fun a -> p (Array.map snd a)
      )
    ) in
  let pskl = atom ((), pskl_reg) in
  make_make_domain_opt speca pskl
;;

(* TODO: Maybe only add a default *)
