(* $Id: fj.mli,v 1.9 2010/10/20 10:56:52 qcarbonn Exp $ *)

(* channels *)
type 'a internal_out_channel = Pervasives.out_channel;;
type 'a internal_in_channel = Pervasives.in_channel;;

type 'a fj_out_channel = 'a option internal_out_channel;;
type 'a fj_in_channel = 'a option internal_in_channel;;


(* skeletons *)
type ('a, 'b) fjtree;;

val map : ('a, 'b) fjtree * int -> ('a, 'b) fjtree;;
val map_vector : ('a, 'b) fjtree * int -> ('a array, 'b array) fjtree;;
val ( ||| ) : ('a, 'b) fjtree -> ('b, 'c) fjtree -> ('a, 'c) fjtree;;
val seq : ('a -> 'b) -> ('a, 'b) fjtree;;


(* Low level primitives *)
type pid = int;;

type ('a, 'b) fork_data =
  | Child of 'a fj_in_channel * 'b fj_out_channel
  | Parent of 'b fj_in_channel * 'a fj_out_channel * int
;;

type ('a, 'b) cfjtree =
  | CFJ of ('a, 'b) cfjtree * ('a, 'b) cfjtree
  | CSeq of 'a fj_in_channel * 'b fj_out_channel
;;

val cpipe : unit -> Pervasives.in_channel * Pervasives.out_channel;;
val receive_in_tree : ('a, 'b) cfjtree -> 'b Streams.t;;
val send_in_tree : ('a, 'b) cfjtree -> 'a Streams.t -> unit;;
val xparfun : ('a, 'b) cfjtree -> 'a Streams.t -> pid list -> 'b Streams.t * pid list;;
val run_function : 'a fj_in_channel -> ('a Streams.t -> 'b Streams.t) -> 'b fj_out_channel -> unit;;
val start_tree : ('a, 'b) fjtree -> pid list -> ('a, 'b) cfjtree * pid list;;

val killall : pid list -> unit;;

val parfun : ('a, 'b) fjtree -> 'a Streams.t -> 'b Streams.t;;
